from typing import Tuple
import math

from PyQt5 import QtCore
from PyQt5.QtCore import QObject

import pyaudio
from pyaudio import PyAudio


__all__ = ["AudioGenerator"]


class AudioGenerator(QObject):
    bitrate = 16000
    frequency = (700, 1500)
    frames_per_buffer = 512

    def __init__(
            self,
            score: float = 0.0,
            score_threshold: float = 0.0,
            play_sound: bool = True,
            parent: QObject = None
    ):
        super().__init__(parent=parent)

        self.__score = float(score)
        self.__score_threshold = float(score_threshold)
        self.__mute_sound = not play_sound

        self.__previous_was_muted = True
        self.__previous_phase = 0.0

        self.__backend = PyAudio()
        self.__stream = self.__backend.open(
            format=self.__backend.get_format_from_width(1),
            channels=1,
            rate=self.bitrate,
            output=True,
            frames_per_buffer=self.frames_per_buffer,
            stream_callback=self.__continue_playing
        )

    def __del__(self):
        self.__stream.stop_stream()
        self.__stream.close()
        self.__backend.terminate()
        del self.__backend

    def __refresh(self):
        return
        """
        if self.__mute_sound or self.__score < self.__score_threshold:
            if self.__stream is not None:
                self.__stream.close()
                self.__stream = None
            #return
            #self.__stream.stop_stream()
            # print("stop stream")
        else:
            #self.__stream.start_stream()
            self.__stream = self.__backend.open(
                format=self.__backend.get_format_from_width(1),
                channels=1,
                rate=self.bitrate,
                output=True,
                frames_per_buffer=self.frames_per_buffer,
                stream_callback=self.__continue_playing
            )
            # print("start stream")
        """

    def get_score(self) -> float:
        return self.__score

    @QtCore.pyqtSlot(float)
    def set_score(self, score):
        self.__score = float(score)
        self.__refresh()

    score = property(get_score, set_score)

    def get_score_threshold(self):
        return self.__score_threshold

    def set_score_threshold(self, value):
        self.__score_threshold = value
        self.__refresh()

    score_threshold = property(get_score_threshold, set_score_threshold)

    @QtCore.pyqtSlot()
    def start(self):
        self.__mute_sound = False
        self.__refresh()

    @QtCore.pyqtSlot()
    def pause(self):
        self.__mute_sound = True
        self.__refresh()

    def __continue_playing(self, id_data: None, frame_count: int, time_info: dict, status_flags: int) \
            -> Tuple[bytes, int]:
        score = self.__score
        k = math.pi * self.__get_frequency(score) / self.bitrate

        if self.__mute_sound or score < self.__score_threshold:
            if self.__previous_was_muted:
                wave_data = bytes(frame_count)
            else:
                wave_data = bytearray(frame_count)

                value_min = int(math.sin(self.__previous_phase) * 127.0 + 128.0)
                for frame in range(frame_count):
                    value = int(math.sin(k * frame + self.__previous_phase) * 127.0 + 128.0)
                    if value_min <= value < 0 or 0 < value <= value_min:
                        value_min = value
                        wave_data[frame] = value
                    else:
                        break

                wave_data = bytes(wave_data)
                self.__previous_was_muted = True
                self.__previous_phase = 0.0
        else:
            wave_data = bytes(
                int(math.sin(k * frame + self.__previous_phase) * 127.0 + 128.0) for frame in range(frame_count)
            )

            self.__previous_was_muted = False
            self.__previous_phase += k * frame_count

        return wave_data, pyaudio.paContinue

    def __get_frequency(self, score: float) -> float:
        return _scale_frequency_linear(score, self.__score_threshold, self.frequency[0], self.frequency[1])


def _scale_frequency_linear(score: float, score_threshold: float, frequency_lower: float, frequency_upper: float) \
        -> float:
    """Frequency as linear function of score, if score >= score_threshold. Else 0."""

    if score > score_threshold:
        return float((score - score_threshold) * (frequency_upper - frequency_lower) + frequency_lower)
    else:
        return 0.0
